@@ -723,6 +723,32 @@ eval_groups(int local, uint32_t nodeid, int nodeStatus)
 			continue;
 		}
 
+		/* Mark the service as stopped if applicable */
+		if ((svcStatus.rs_owner == nodeid && !nodeStatus) &&
+		    (svcStatus.rs_state == RG_STATE_STARTED ||
+		     svcStatus.rs_state == RG_STATE_RECOVER ||
+		     svcStatus.rs_state == RG_STATE_STARTING ||
+		     svcStatus.rs_state == RG_STATE_STOPPING )) {
+
+			logt_print(LOG_DEBUG,
+				   "Marking %s on down member %d as stopped",
+				   svcName, nodeid);
+
+			svcStatus.rs_last_owner = svcStatus.rs_owner;
+			svcStatus.rs_state = RG_STATE_STOPPED;
+			svcStatus.rs_owner = 0;
+			svcStatus.rs_transition = (uint64_t)time(NULL);
+			svcStatus.rs_flags = 0;
+
+			if (set_rg_state(svcName, &svcStatus) != 0) {
+				logt_print(LOG_ERR, "Failed to update state"
+					   " of %s during recovery; cannot "
+					   "fail over", svcName);
+				rg_unlock(&lockp);
+				continue;
+			}
+		}
+
 		rg_unlock(&lockp);
 
 		if (svcStatus.rs_owner == 0)
@@ -975,6 +1001,9 @@ group_op(const char *groupname, int op)
 	case RG_CONDSTART:
 		ret = res_condstart(&_tree, res, NULL);
 		break;
+	case RG_CONVALESCE:
+		ret = res_convalesce(&_tree, res, NULL);
+		break;
 	}
 	pthread_rwlock_unlock(&resource_lock);
 
@@ -1273,13 +1302,21 @@ svc_exists(const char *svcname)
 }
 
 
-void
+/*
+ * Perform an operation on all resources groups.
+ *
+ * Returns the number of requests queued.  This value is
+ * only used during shutdown, where we queue RG_STOP_EXITING
+ * only for services we have running locally as an optimization.
+ */
+int
 rg_doall(int request, int block,
 	 const char __attribute__ ((unused)) *debugfmt)
 {
 	resource_node_t *curr;
 	rg_state_t svcblk;
 	char rg[64];
+	int queued = 0;
 
 	pthread_rwlock_rdlock(&resource_lock);
 	list_do(&_tree, curr) {
@@ -1304,6 +1341,7 @@ rg_doall(int request, int block,
 
 		rt_enqueue_request(rg, request, NULL, 0,
 				   0, 0, 0);
+		++queued;
 	} while (!list_done(&_tree, curr));
 
 	pthread_rwlock_unlock(&resource_lock);
@@ -1313,6 +1351,7 @@ rg_doall(int request, int block,
 	   other rgmanagers to complete. */
 	if (block) 
 		rg_wait_threads();
+	return queued;
 }
 
 
@@ -1670,8 +1709,6 @@ init_resource_groups(int reconfigure, int do_init)
 		free(val);
 	}
 
-	/* Wait for any pending requests */
-	rg_wait_threads();
 	/* Block operations that would break during configuration
 	   changes */
 	rg_clear_initialized(FL_CONFIG);
@@ -1686,6 +1723,8 @@ init_resource_groups(int reconfigure, int do_init)
 		return -1;
 	}
 
+	load_resource_defaults(fd, &rulelist);
+
 	if (build_resource_tree(fd, &tree, &rulelist, &reslist) != 0) {
 		logt_print(LOG_CRIT, "#7: Error building resource tree\n");
 		destroy_resource_tree(&tree);
